/// <reference path="../types/all.d.ts" />
"use strict";
var fs = require('fs');
var fse = require('fs-extra');
var path = require('path');
var debug = require('debug');
var ts = require('typescript');
var diffing_broccoli_plugin_1 = require('./diffing-broccoli-plugin');
var FS_OPTS = {
    encoding: 'utf-8'
};
// Sub-directory where the @internal typing files (.d.ts) are stored
exports.INTERNAL_TYPINGS_PATH = 'internal_typings';
// Monkey patch the TS compiler to be able to re-emit files with @internal symbols
var tsEmitInternal = false;
var originalEmitFiles = ts.emitFiles;
ts.emitFiles = function (resolver, host, targetSourceFile) {
    if (tsEmitInternal) {
        var orignalgetCompilerOptions_1 = host.getCompilerOptions;
        host.getCompilerOptions = function () {
            var options = clone(orignalgetCompilerOptions_1.call(host));
            options.stripInternal = false;
            options.outDir = options.outDir + "/" + exports.INTERNAL_TYPINGS_PATH;
            return options;
        };
    }
    return originalEmitFiles(resolver, host, targetSourceFile);
};
/**
 * Broccoli plugin that implements incremental Typescript compiler.
 *
 * It instantiates a typescript compiler instance that keeps all the state about the project and
 * can re-emit only the files that actually changed.
 *
 * Limitations: only files that map directly to the changed source file via naming conventions are
 * re-emitted. This primarily affects code that uses `const enum`s, because changing the enum value
 * requires global emit, which can affect many files.
 */
var DiffingTSCompiler = (function () {
    function DiffingTSCompiler(inputPath, cachePath, options) {
        this.inputPath = inputPath;
        this.cachePath = cachePath;
        this.options = options;
        this.fileRegistry = Object.create(null);
        this.firstRun = true;
        this.previousRunFailed = false;
        // Whether to generate the @internal typing files (they are only generated when `stripInternal` is
        // true)
        this.genInternalTypings = false;
        if (options && options.rootFilePaths) {
            this.rootFilePaths = options.rootFilePaths.splice(0);
        }
        else {
            this.rootFilePaths = [];
        }
        if (options && options.includeExtensions) {
            DiffingTSCompiler.includeExtensions = options.includeExtensions;
        }
        if (options && options.internalTypings) {
            this.genInternalTypings = true;
        }
        this.tsOpts = (options && options.tsOptions) || {};
        if (this.tsOpts.stripInternal === false) {
            // @internal are included in the generated .d.ts, do not generate them separately
            this.genInternalTypings = false;
        }
        this.tsOpts.rootDir = inputPath;
        this.tsOpts.outDir = this.cachePath;
        if (this.rootFilePaths && this.rootFilePaths.length) {
            this._debug("CustomLanguageServiceHost rootFilePaths " + this.rootFilePaths.join(";"));
        }
        this._debug("CustomLanguageServiceHost inputPath " + this.inputPath);
        var localTypesFolder = options.localTypesFolder || process.cwd() + "/local-types";
        this.tsServiceHost = new CustomLanguageServiceHost(this.tsOpts, this.rootFilePaths, this.fileRegistry, this.inputPath, localTypesFolder);
        this.tsService = ts.createLanguageService(this.tsServiceHost, ts.createDocumentRegistry());
    }
    DiffingTSCompiler.prototype._debug = function (message) {
        debug('broccoli-typify')("broccoli-typescript " + message);
    };
    DiffingTSCompiler.prototype.rebuild = function (treeDiff) {
        var _this = this;
        var pathsToEmit = [];
        var pathsWithErrors = [];
        var errorMessages = [];
        treeDiff.addedPaths.concat(treeDiff.changedPaths).forEach(function (tsFilePath) {
            if (!_this.fileRegistry[tsFilePath]) {
                _this.fileRegistry[tsFilePath] = { version: 0 };
                _this.rootFilePaths.push(tsFilePath);
            }
            else {
                _this.fileRegistry[tsFilePath].version++;
            }
            pathsToEmit.push(path.join(_this.inputPath, tsFilePath));
        });
        treeDiff.removedPaths.forEach(function (tsFilePath) {
            console.log('removing outputs for', tsFilePath);
            _this.rootFilePaths.splice(_this.rootFilePaths.indexOf(tsFilePath), 1);
            _this.fileRegistry[tsFilePath] = null;
            _this.removeOutputFor(tsFilePath);
        });
        if (this.firstRun) {
            this.firstRun = false;
            this.doFullBuild();
        }
        else {
            var program_1 = this.tsService.getProgram();
            tsEmitInternal = false;
            pathsToEmit.forEach(function (tsFilePath) {
                var output = _this.tsService.getEmitOutput(tsFilePath);
                if (output.emitSkipped) {
                    var errorFound = _this.collectErrors(tsFilePath);
                    if (errorFound) {
                        pathsWithErrors.push(tsFilePath);
                        errorMessages.push(errorFound);
                    }
                }
                else {
                    output.outputFiles.forEach(function (o) {
                        var destDirPath = path.dirname(o.name);
                        fse.mkdirsSync(destDirPath);
                        fs.writeFileSync(o.name, o.text, FS_OPTS);
                        if (endsWith(o.name, '.d.ts')) {
                            var sourceFile = program_1.getSourceFile(tsFilePath);
                        }
                    });
                }
            });
            if (pathsWithErrors.length) {
                this.previousRunFailed = true;
                var error = new Error('Typescript found the following errors:\n' + errorMessages.join('\n'));
                error['showStack'] = false;
                throw error;
            }
            else if (this.previousRunFailed) {
                this.doFullBuild();
            }
            else if (this.genInternalTypings) {
                // serialize the .d.ts files containing @internal symbols
                tsEmitInternal = true;
                pathsToEmit.forEach(function (tsFilePath) {
                    var output = _this.tsService.getEmitOutput(tsFilePath);
                    if (!output.emitSkipped) {
                        output.outputFiles.forEach(function (o) {
                            if (endsWith(o.name, '.d.ts')) {
                                var destDirPath = path.dirname(o.name);
                                fse.mkdirsSync(destDirPath);
                                fs.writeFileSync(o.name, o.text, FS_OPTS);
                            }
                        });
                    }
                });
                tsEmitInternal = false;
            }
        }
    };
    DiffingTSCompiler.prototype.collectErrors = function (tsFilePath) {
        var allDiagnostics = this.tsService.getCompilerOptionsDiagnostics()
            .concat(this.tsService.getSyntacticDiagnostics(tsFilePath))
            .concat(this.tsService.getSemanticDiagnostics(tsFilePath));
        var errors = [];
        allDiagnostics.forEach(function (diagnostic) {
            var message = ts.flattenDiagnosticMessageText(diagnostic.messageText, '\n');
            if (diagnostic.file) {
                var _a = diagnostic.file.getLineAndCharacterOfPosition(diagnostic.start), line = _a.line, character = _a.character;
                errors.push("  " + diagnostic.file.fileName + " (" + (line + 1) + "," + (character + 1) + "): " + message);
            }
            else {
                errors.push("  Error: " + message);
            }
        });
        if (errors.length) {
            return errors.join('\n');
        }
    };
    DiffingTSCompiler.prototype.doFullBuild = function () {
        var _this = this;
        var program = this.tsService.getProgram();
        var typeChecker = program.getTypeChecker();
        var diagnostics = [];
        tsEmitInternal = false;
        var emitResult = program.emit(undefined, function (absoluteFilePath, fileContent) {
            fse.mkdirsSync(path.dirname(absoluteFilePath));
            fs.writeFileSync(absoluteFilePath, fileContent, FS_OPTS);
            if (endsWith(absoluteFilePath, '.d.ts')) {
                // TODO: Use sourceFile from the callback if
                //   https://github.com/Microsoft/TypeScript/issues/7438
                // is taken
                var originalFile = absoluteFilePath.replace(_this.tsOpts.outDir, _this.tsOpts.rootDir)
                    .replace(/\.d\.ts$/, '.ts');
                var sourceFile = program.getSourceFile(originalFile);
            }
        });
        if (this.genInternalTypings) {
            // serialize the .d.ts files containing @internal symbols
            tsEmitInternal = true;
            program.emit(undefined, function (absoluteFilePath, fileContent) {
                if (endsWith(absoluteFilePath, '.d.ts')) {
                    fse.mkdirsSync(path.dirname(absoluteFilePath));
                    fs.writeFileSync(absoluteFilePath, fileContent, FS_OPTS);
                }
            });
            tsEmitInternal = false;
        }
        if (emitResult.emitSkipped) {
            var allDiagnostics = ts.getPreEmitDiagnostics(program).concat(emitResult.diagnostics);
            var errorMessages_1 = [];
            allDiagnostics.forEach(function (diagnostic) {
                var pos = '';
                if (diagnostic.file) {
                    var _a = diagnostic.file.getLineAndCharacterOfPosition(diagnostic.start), line = _a.line, character = _a.character;
                    pos = diagnostic.file.fileName + " (" + (line + 1) + ", " + (character + 1) + "): ";
                }
                var message = ts.flattenDiagnosticMessageText(diagnostic.messageText, '\n');
                errorMessages_1.push("  " + pos + message);
            });
            if (errorMessages_1.length) {
                this.previousRunFailed = true;
                var error = new Error('Typescript found the following errors:\n' + errorMessages_1.join('\n'));
                error['showStack'] = false;
                throw error;
            }
            else {
                this.previousRunFailed = false;
            }
        }
    };
    DiffingTSCompiler.prototype.removeOutputFor = function (tsFilePath) {
        var absoluteJsFilePath = path.join(this.cachePath, tsFilePath.replace(/\.ts$/, '.js'));
        var absoluteMapFilePath = path.join(this.cachePath, tsFilePath.replace(/.ts$/, '.js.map'));
        var absoluteDtsFilePath = path.join(this.cachePath, tsFilePath.replace(/\.ts$/, '.d.ts'));
        if (fs.existsSync(absoluteJsFilePath)) {
            fs.unlinkSync(absoluteJsFilePath);
            if (fs.existsSync(absoluteMapFilePath)) {
                // source map could be inline or not generated
                fs.unlinkSync(absoluteMapFilePath);
            }
            fs.unlinkSync(absoluteDtsFilePath);
        }
    };
    DiffingTSCompiler.includeExtensions = ['.ts', '.js'];
    return DiffingTSCompiler;
}());
function fileExists(fileName) {
    return ts.sys.fileExists(fileName);
}
function readFile(fileName) {
    return ts.sys.readFile(fileName);
}
var CustomLanguageServiceHost = (function () {
    function CustomLanguageServiceHost(compilerOptions, fileNames, fileRegistry, treeInputPath, localTypesFolder) {
        this.compilerOptions = compilerOptions;
        this.fileNames = fileNames;
        this.fileRegistry = fileRegistry;
        this.treeInputPath = treeInputPath;
        this.localTypesFolder = localTypesFolder;
        this.currentDirectory = process.cwd();
        this.defaultLibFilePath = ts.getDefaultLibFilePath(compilerOptions).replace(/\\/g, '/');
    }
    CustomLanguageServiceHost.prototype.getScriptFileNames = function () {
        var _this = this;
        return this.fileNames.map(function (f) { return path.join(_this.treeInputPath, f); });
    };
    CustomLanguageServiceHost.prototype.getScriptVersion = function (fileName) {
        if (startsWith(fileName, this.treeInputPath)) {
            var key = fileName.substr(this.treeInputPath.length + 1);
            return this.fileRegistry[key] && this.fileRegistry[key].version.toString();
        }
    };
    CustomLanguageServiceHost.prototype.getScriptSnapshot = function (tsFilePath) {
        // TypeScript seems to request lots of bogus paths during import path lookup and resolution,
        // so we we just return undefined when the path is not correct.
        // Ensure it is in the input tree, an imported @type files or lib/*d.ts file.
        if (!startsWith(tsFilePath, this.treeInputPath)
            && tsFilePath.indexOf('/node_modules/@types/') === -1
            && tsFilePath.indexOf('/node_modules/at-types') === -1
            && tsFilePath.indexOf(this.localTypesFolder) === -1
            && !tsFilePath.match(/\/lib(\..*)*.d\.ts$/)) {
            if (fs.existsSync(tsFilePath)) {
                console.log('Rejecting', tsFilePath, '. File is not in the input tree.');
            }
            return undefined;
        }
        // Ensure it exists
        if (!fs.existsSync(tsFilePath)) {
            return undefined;
        }
        return ts.ScriptSnapshot.fromString(fs.readFileSync(tsFilePath, FS_OPTS));
    };
    CustomLanguageServiceHost.prototype.getCurrentDirectory = function () { return this.currentDirectory; };
    CustomLanguageServiceHost.prototype.getCompilationSettings = function () { return this.compilerOptions; };
    CustomLanguageServiceHost.prototype.getDefaultLibFileName = function (options) {
        // ignore options argument, options should not change during the lifetime of the plugin
        return this.defaultLibFilePath;
    };
    CustomLanguageServiceHost.prototype.resolveModuleNames = function (moduleNames, containingFile) {
        var _this = this;
        return moduleNames.map(function (name) {
            // first try the default resolution
            var result = ts.resolveModuleName(name, containingFile, _this.compilerOptions, { fileExists: fileExists, readFile: readFile });
            if (result.resolvedModule) {
                return result.resolvedModule;
            }
            var candidatePaths = [];
            if (name === 'ember') {
                // custom repo for ember resolution while we stabilize the type definition
                candidatePaths.push(_this.currentDirectory + "/node_modules/at-types-ember/index.d.ts");
            }
            if (name.match(/\/config\/environment$/)) {
                candidatePaths.push(_this.localTypesFolder + "/ember-config-environment.d.ts");
            }
            else {
                // resolve npm: modules as loaded with ember-browserify.
                // the end goal is to have all the types coming from npm @types,
                // however we support a local-types for development.
                var module_1 = (name.indexOf('npm:') === 0) ? name.split(':')[1] : name;
                candidatePaths.push(_this.localTypesFolder + "/" + module_1 + "/index.d.ts");
                candidatePaths.push(_this.currentDirectory + "/node_modules/@types/" + module_1 + "/index.d.ts");
            }
            for (var i = 0; i < candidatePaths.length; i++) {
                if (fs.existsSync(candidatePaths[i])) {
                    return {
                        resolvedFileName: candidatePaths[i],
                        isExternalLibraryImport: true
                    };
                }
            }
            debug('broccoli-typify')("resolveModuleNames skipping module '" + name + "'");
            return undefined;
        });
    };
    return CustomLanguageServiceHost;
}());
exports.Compiler = diffing_broccoli_plugin_1.wrapDiffingPlugin(DiffingTSCompiler);
function clone(object) {
    var result = {};
    for (var id in object) {
        result[id] = object[id];
    }
    return result;
}
function startsWith(str, substring) {
    return str.substring(0, substring.length) === substring;
}
function endsWith(str, substring) {
    return str.indexOf(substring, str.length - substring.length) !== -1;
}
